"""Core module for metadata extraction job settings."""

import argparse
import json
import logging
from pydantic_settings import (
    BaseSettings,
    EnvSettingsSource,
    InitSettingsSource,
    JsonConfigSettingsSource,
    PydanticBaseSettingsSource,
)
from pydantic import Field
from typing import Optional, Union, List, Type, Tuple
from pathlib import Path


class BaseExtractor:
    """Parent class for metadata extractors."""

    def _extract(self):
        """Load and extract metadata from input source."""
        raise NotImplementedError("This method should be implemented by subclasses.")

    def run_job(self):
        """Run the extraction job."""
        raise NotImplementedError("This method should be implemented by subclasses.")

    def write(self) -> None:
        """Save extraction results to a standardized JSON file."""
        if not hasattr(self, "metadata"):
            raise ValueError("No metadata found. Please run the job first.")

        job_settings = getattr(self, "job_settings", None)
        if job_settings is None or getattr(job_settings, "output_directory", None) is None:
            raise ValueError("No output directory specified in job settings.")

        job_settings.output_directory.mkdir(parents=True, exist_ok=True)

        # Generate filename from the module's parent directory name
        # e.g., aind_metadata_extractor.mesoscope.extractor -> "mesoscope"
        module_path = self.__class__.__module__
        module_parts = module_path.split(".")
        if len(module_parts) >= 2:
            folder_name = module_parts[-2]  # Get the parent folder name
        else:
            raise ValueError("Cannot determine folder name from module path.")

        output_path = job_settings.output_directory / f"{folder_name}.json"

        metadata = getattr(self, "metadata")
        with open(output_path, "w") as f:
            # Handle both pydantic models and plain dicts
            if hasattr(metadata, "model_dump"):
                json.dump(metadata.model_dump(), f, indent=4)
            else:
                json.dump(metadata, f, default=str, indent=4)

        logging.info(f"Metadata written to {output_path}")


class BaseJobSettings(BaseSettings):
    """Parent class for generating settings from a config file."""

    job_settings_name: str = Field(
        ...,
        description=("Literal name for job settings to make serialized class distinct."),
    )
    input_source: Optional[Union[Path, str, List[str], List[Path]]] = Field(
        default=None,
        description=("Location or locations of data sources to parse for metadata."),
    )
    output_directory: Optional[Union[Path, str]] = Field(
        default=None,
        description=("Location to metadata file data to. None to return object."),
    )

    user_settings_config_file: Optional[Union[Path, str]] = Field(
        default=None,
        repr=False,
        description="Optionally pull settings from a local config file.",
    )

    class Config:
        """Pydantic config to exclude field from displaying"""

        extra = "allow"
        exclude = {"user_settings_config_file"}

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: InitSettingsSource,
        env_settings: EnvSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        """
        Customize the order of settings sources, including JSON file.
        """
        config_file = init_settings.init_kwargs.get("user_settings_config_file")
        sources = [init_settings, env_settings]

        if isinstance(config_file, str):
            config_file = Path(config_file)

        if config_file and config_file.is_file():
            try:
                sources.append(JsonConfigSettingsSource(settings_cls, config_file))
            except Exception as e:
                logging.warning(f"Failed to load JSON config file {config_file}: {e}")
                raise

        return tuple(sources)

    @classmethod
    def from_args(cls, args: list):
        """
        Adds ability to construct settings from a list of arguments.
        Parameters
        ----------
        args : list
        A list of command line arguments to parse.
        """

        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-j",
            "--job-settings",
            required=True,
            type=str,
            help=(r"""
                Custom settings defined by the user defined as a json
                 string. For example: -j
                 '{
                 "input_source":"/directory/to/read/from",
                 "output_directory":"/directory/to/write/to",
                 "job_settings_name": "Bergamo"}'
                """),
        )
        job_args = parser.parse_args(args)
        job_settings_from_args = cls.model_validate_json(job_args.job_settings)
        return job_settings_from_args



"""Constants for the naming utils of metadata mapper"""

import re

INT_NULL = -99

default_stimulus_renames = {
    "": "spontaneous",
    "natural_movie_1": "natural_movie_one",
    "natural_movie_3": "natural_movie_three",
    "Natural Images": "natural_scenes",
    "flash_250ms": "flashes",
    "gabor_20_deg_250ms": "gabors",
    "drifting_gratings": "drifting_gratings",
    "static_gratings": "static_gratings",
    "contrast_response": "drifting_gratings_contrast",
    "Natural_Images_Shuffled": "natural_scenes_shuffled",
    "Natural_Images_Sequential": "natural_scenes_sequential",
    "natural_movie_1_more_repeats": "natural_movie_one",
    "natural_movie_shuffled": "natural_movie_one_shuffled",
    "motion_stimulus": "dot_motion",
    "drifting_gratings_more_repeats": "drifting_gratings_75_repeats",
    "signal_noise_test_0_200_repeats": "test_movie_one",
    "signal_noise_test_0": "test_movie_one",
    "signal_noise_test_1": "test_movie_two",
    "signal_noise_session_1": "dense_movie_one",
    "signal_noise_session_2": "dense_movie_two",
    "signal_noise_session_3": "dense_movie_three",
    "signal_noise_session_4": "dense_movie_four",
    "signal_noise_session_5": "dense_movie_five",
    "signal_noise_session_6": "dense_movie_six",
}


default_column_renames = {
    "Contrast": "contrast",
    "Ori": "orientation",
    "Oris": "orientation",
    "SF": "spatial_frequency",
    "TF": "temporal_frequency",
    "Size": "size",
    "Phase": "phase",
    "Color": "color",
    "Image": "frame",
    "Pos_x": "x_position",
    "Pos_y": "y_position",
}


GABOR_DIAMETER_RE = re.compile(r"gabor_(\d*\.{0,1}\d*)_{0,1}deg(?:_\d+ms){0,1}")

GENERIC_MOVIE_RE = re.compile(
    r"natural_movie_"
    + r"(?P<number>\d+|one|two|three|four|five|six|seven|eight|nine)"
    + r"(_shuffled){0,1}(_more_repeats){0,1}"
)
DIGIT_NAMES = {
    "1": "one",
    "2": "two",
    "3": "three",
    "4": "four",
    "5": "five",
    "6": "six",
    "7": "seven",
    "8": "eight",
    "9": "nine",
}
SHUFFLED_MOVIE_RE = re.compile(r"natural_movie_shuffled")
NUMERAL_RE = re.compile(r"(?P<number>\d+)")



"""Utils to process pkl files"""

import pickle

import numpy as np
import pandas as pd


def load_pkl(path):
    """
    Loads a pkl stim file

    Parameters
    ----------
    path : str
        Path to pkl file.

    Returns
    -------
    data : dict
        pkl file.
    """
    data = pd.read_pickle(path)
    return data


def load_img_pkl(pstream):
    """
    Loads a pkl stim file

    Parameters
    ----------
    pstream : str
        image pkl file.

    """
    return pickle.load(pstream, encoding="Latin-1")


def get_stimuli(pkl, is_behavior=False):
    """
    Returns the stimuli from a pkl file

    Parameters
    ----------
    pkl : dict
        pkl file.

    """

    if is_behavior:
        return pkl["items"]["behavior"]["items"]
    return pkl["stimuli"]


def get_fps(pkl):
    """
    Returns the fps from a pkl file

    Parameters
    ----------
    pkl : dict
        pkl file.

    Returns
    -------
    data: int
        fps.

    """
    if not pkl.get("fps"):
        fps = round(1 / np.mean(pkl["items"]["behavior"]["intervalsms"]) * 0.001, 2)
    else:
        fps = pkl["fps"]
    return fps


def get_pre_blank_sec(pkl):
    """
    Returns the pre_blank_sec from a pkl file

    Parameters
    ----------

    pkl : dict
        pkl file.

    Returns
    -------
    data: int
        pre_blank_sec.

    """
    return pkl["pre_blank_sec"]


def angular_wheel_velocity(pkl):
    """
    Returns the wheel velocity from a pkl file

    Parameters
    ----------
    pkl : dict
        pkl file.

    Returns
    -------
    data: int
        fps * wheel rotation speed

    """
    return get_fps(pkl) * get_angular_wheel_rotation(pkl)


def get_stage(pkl):
    """
    Returns the stage from a pkl file

    Parameters
    ----------
    pkl : dict
        pkl file.

    Returns
    -------
    data: str
        stage name
    """
    if "stage" in pkl:
        return pkl["stage"]
    elif "items" in pkl:
        return pkl["items"]["behavior"]["cl_params"]["stage"]


def get_angular_wheel_rotation(pkl):
    """
    Returns the wheel rotation from a pkl file

    Parameters
    ----------
    pkl : dict
        pkl file.

    Returns
    -------
    data: int
        wheel rotation speed

    """
    return get_running_array(pkl, "dx")


def vsig(pkl):
    """
    Returns the vsig from a pkl file

    Parameters
    ----------
    pkl : dict
        pkl file.

    Returns
    -------
    data: int
        vsig

    """
    return get_running_array(pkl, "vsig")


def vin(pkl):
    """
    Returns the voltage in from a pkl file

    Parameters
    ----------

    pkl : dict
        pkl file.

    Returns
    -------
    data: vin
        voltage in

    """
    return get_running_array(pkl, "vin")


def get_running_array(pkl, key):
    """
    Returns an running array from a pkl file

    Parameters
    ----------
    pkl : dict
        pkl file.
    key : str
        key to extract from pkl file.

    Returns
    -------
    data: array
        running array

    """
    try:
        result = pkl["items"]["foraging"]["encoders"][0][key]
    except (KeyError, IndexError):
        try:
            result = pkl[key]
        except KeyError:
            raise KeyError(f"unable to extract {key} from this stimulus pickle")

    return np.array(result)



"""
File containing Camstim class
"""

import functools
from datetime import timedelta
from pathlib import Path
from typing import Optional

import pandas as pd
from pydantic import BaseModel

from aind_metadata_extractor.utils.camstim_sync import (
    behavior_utils,
    constants,
    pkl_utils as pkl,
    stim_utils,
    sync_utils,
    naming_utils,
)


class CamstimSettings(BaseModel):
    """Camstim settings for extracting stimulus epochs"""

    sessions_root: Optional[Path] = None
    opto_conditions_map: Optional[dict] = None
    overwrite_tables: bool = False
    input_source: Path
    output_directory: Optional[Path]
    session_id: str
    subject_id: str


class Camstim:
    """
    Methods used to extract stimulus epochs
    """

    def __init__(
        self,
        camstim_settings: CamstimSettings,
    ) -> None:
        """
        Determine needed input filepaths from np-exp and lims, get session
        start and end times from sync file, write stim tables and extract
        epochs from stim tables. If 'overwrite_tables' is not given as True,
        in the json settings and an existing stim table exists, a new one
        won't be written. opto_conditions_map may be given in the json
        settings to specify the different laser states for this experiment.
        Otherwise, the default is used from naming_utils.
        """
        self.sync_path = None
        self.sync_data = None
        self.camstim_settings = camstim_settings
        self.input_source = Path(self.camstim_settings.input_source)
        session_id = self.camstim_settings.session_id
        self.pkl_path = next(self.input_source.rglob("*.pkl"))
        if not self.camstim_settings.output_directory.is_dir():
            self.camstim_settings.output_directory.mkdir(parents=True)
        self.stim_table_path = self.camstim_settings.output_directory / f"{session_id}_stim_table.csv"
        self.vsync_table_path = self.camstim_settings.output_directory / f"{session_id}_vsync_table.csv"
        self.pkl_data = pkl.load_pkl(self.pkl_path)
        self.fps = pkl.get_fps(self.pkl_data)
        self.stage_name = pkl.get_stage(self.pkl_data)
        self.session_start, self.session_end = self._get_sync_times()
        self.sync_data = sync_utils.load_sync(self.sync_path)
        self.mouse_id = self.camstim_settings.subject_id
        self.session_uuid = self.get_session_uuid()
        self.behavior = self._is_behavior()
        self.session_type = self._get_session_type()

    def _get_session_type(self) -> str:
        """Determine the session type from the pickle data

        Returns
        -------
        str
            session type
        """
        if self.behavior:
            return self.pkl_data["items"]["behavior"]["params"]["stage"]
        else:
            return self.pkl_data["items"]["foraging"]["params"]["stage"]

    def _is_behavior(self) -> bool:
        """Check if the session has behavior data"""
        if self.pkl_data.get("items", {}).get("behavior", None):
            return True
        return False

    def _get_sync_times(self) -> None:
        """Set the sync path
        Returns
        -------
        Path
        """
        self.sync_path = next(self.input_source.glob("*.h5"))
        self.sync_data = sync_utils.load_sync(self.sync_path)
        return sync_utils.get_start_time(self.sync_data), sync_utils.get_stop_time(self.sync_data)

    def build_behavior_table(self) -> None:
        """Builds a behavior table from the stimulus pickle file and writes it
        to a csv file

        Returns
        -------
        None
        """
        timestamps = sync_utils.get_ophys_stimulus_timestamps(self.sync_data, self.pkl_path)
        behavior_table = behavior_utils.from_stimulus_file(self.pkl_path, timestamps)
        behavior_table[0].to_csv(self.stim_table_path, index=False)

    def get_session_uuid(self) -> str:
        """Returns the session uuid from the pickle file"""
        return pkl.load_pkl(self.pkl_path)["session_uuid"]

    def get_stim_table_seconds(self, stim_table_sweeps, frame_times, name_map) -> pd.DataFrame:
        """Builds a stimulus table from the stimulus pickle file, sync file

        Parameters
        ----------
        stim_table_sweeps : pd.DataFrame
            DataFrame containing stimulus information
        frame_times : np.array
            Array containing frame times
        name_map : dict
            Dictionary containing stimulus names

        Returns
        -------
        pd.DataFrame
        """
        stim_table_seconds = stim_utils.convert_frames_to_seconds(stim_table_sweeps, frame_times, self.fps, True)
        stim_table_seconds = naming_utils.collapse_columns(stim_table_seconds)
        stim_table_seconds = naming_utils.drop_empty_columns(stim_table_seconds)
        stim_table_seconds = naming_utils.standardize_movie_numbers(stim_table_seconds)
        stim_table_seconds = naming_utils.add_number_to_shuffled_movie(stim_table_seconds)
        stim_table_seconds = naming_utils.map_stimulus_names(stim_table_seconds, name_map)
        return stim_table_seconds

    def build_stimulus_table(
        self,
        minimum_spontaneous_activity_duration=0.0,
        extract_const_params_from_repr=False,
        drop_const_params=stim_utils.DROP_PARAMS,
        stimulus_name_map=constants.default_stimulus_renames,
        column_name_map=constants.default_column_renames,
        modality="ephys",
    ):
        """
        Builds a stimulus table from the stimulus pickle file, sync file, and
        the given parameters. Writes the table to a csv file.

        Parameters
        ----------
        minimum_spontaneous_activity_duration : float, optional
            Minimum duration of spontaneous activity to be considered a
            separate epoch, by default 0.0
        extract_const_params_from_repr : bool, optional
            Whether to extract constant parameters from the stimulus
            representation, by default False
        drop_const_params : list[str], optional
            List of constant parameters to drop, by default stim.DROP_PARAMS
        stimulus_name_map : dict[str, str], optional
            Map of stimulus names to rename, by default
            naming_utils.default_stimulus_renames
        column_name_map : dict[str, str], optional
            Map of column names to rename, by default
            naming_utils.default_column_renames

        """
        assert not self.behavior, "Can't generate regular stim table from behavior pkl. \
            Use build_behavior_table instead."

        vsync_times = stim_utils.extract_frame_times_from_vsync(self.sync_data)
        if modality == "ephys":
            frame_times = stim_utils.extract_frame_times_from_photodiode(self.sync_data)
            times = [frame_times]

        elif modality == "ophys":
            delay = stim_utils.extract_frame_times_with_delay(self.sync_data)
            frame_times = stim_utils.extract_frame_times_from_vsync(self.sync_data)
            frame_times = frame_times + delay
            times = [frame_times, vsync_times]

        for i, time in enumerate(times):
            minimum_spontaneous_activity_duration = minimum_spontaneous_activity_duration / pkl.get_fps(self.pkl_data)

            stimulus_table = functools.partial(
                stim_utils.build_stimuluswise_table,
                seconds_to_frames=stim_utils.seconds_to_frames,
                extract_const_params_from_repr=extract_const_params_from_repr,
                drop_const_params=drop_const_params,
            )

            spon_table = functools.partial(
                stim_utils.make_spontaneous_activity_tables,
                duration_threshold=minimum_spontaneous_activity_duration,
            )

            stimuli = pkl.get_stimuli(self.pkl_data)
            stimuli = stim_utils.extract_blocks_from_stim(stimuli)
            stim_table_sweeps = stim_utils.create_stim_table(self.pkl_data, stimuli, stimulus_table, spon_table)

            stim_table_seconds = self.get_stim_table_seconds(stim_table_sweeps, time, stimulus_name_map)
            stim_table_final = naming_utils.map_column_names(stim_table_seconds, column_name_map, ignore_case=False)
            if i == 0:
                stim_table_final.to_csv(self.stim_table_path, index=False)
            else:
                stim_table_final.to_csv(self.vsync_table_path, index=False)

    def _summarize_epoch_params(
        self,
        stim_table: pd.DataFrame,
        current_epoch: list,
        start_idx: int,
        end_idx: int,
    ):
        """
        This fills in the current_epoch tuple with the set of parameters
        that exist between start_idx and end_idx
        """
        for column in stim_table:
            if column not in (
                "start_time",
                "stop_time",
                "stim_name",
                "stim_type",
                "start_frame",
                "end_frame",
                "frame",
                "duration",
                "image_set",
                "stim_block",
                "flashes_since_change",
                "image_index",
                "is_change",
                "omitted",
            ):
                param_set = set(stim_table[column][start_idx:end_idx].dropna())
                if len(param_set) > 1000:
                    current_epoch[3][column] = ["Error: over 1000 values"]
                elif param_set:
                    current_epoch[3][column] = param_set

    def extract_stim_epochs(self, stim_table: pd.DataFrame) -> list[list[str, int, int, dict, set]]:
        """
        Returns a list of stimulus epochs, where an epoch takes the form
        (name, start, stop, params_dict, template names). Iterates over the
        stimulus epochs table, identifying epochs based on when the
        'stim_name' field of the table changes.

        For each epoch, every unknown column (not start_time, stop_time,
        stim_name, stim_type, or frame) are listed as parameters, and the set
        of values for that column are listed as parameter values.
        """
        epochs = []

        current_epoch = [None, 0.0, 0.0, {}, set()]
        epoch_start_idx = 0
        for current_idx, row in stim_table.iterrows():
            if row["stim_name"] == "spontaneous":
                continue
            if row["stim_name"] != current_epoch[0]:
                self._summarize_epoch_params(stim_table, current_epoch, epoch_start_idx, current_idx)
                epochs.append(current_epoch)
                epoch_start_idx = current_idx
                current_epoch = [
                    row["stim_name"],
                    row["start_time"],
                    row["stop_time"],
                    {},
                    set(),
                ]
            else:
                current_epoch[2] = row["stop_time"]

            stim_name = row.get("stim_name", "") or ""
            image_set = row.get("image_set", "")
            if pd.notnull(image_set) and image_set:  # Check both not null and not empty
                stim_name = image_set

            if "image" in stim_name.lower() or "movie" in stim_name.lower():
                current_epoch[4].add(row["stim_name"])

        # Process the final epoch
        if current_epoch[0] is not None:
            self._summarize_epoch_params(stim_table, current_epoch, epoch_start_idx, len(stim_table))
            epochs.append(current_epoch)

        return epochs[1:]

    def epochs_from_stim_table(self) -> list[dict]:
        """
        From the stimulus epochs table, return a list of schema stimulus
        epochs representing the various periods of stimulus from the session.
        Also include the camstim version from pickle file and stimulus script
        used from mtrain.
        """

        software_obj = dict(
            name="camstim",
            version="1.0",
            url="https://eng-gitlab.corp.alleninstitute.org/braintv/camstim",
        )

        script_obj = dict(name=self.stage_name, version="1.0")

        schema_epochs = []
        for (
            epoch_name,
            epoch_start,
            epoch_end,
            stim_params,
            stim_template_names,
        ) in self.extract_stim_epochs(pd.read_csv(self.stim_table_path)):
            params_obj = dict(
                stimulus_name=epoch_name,
                stimulus_parameters=stim_params,
                stimulus_template_name=stim_template_names,
            )

            epoch_obj = dict(
                stimulus_start_time=self.session_start + timedelta(seconds=epoch_start),
                stimulus_end_time=self.session_start + timedelta(seconds=epoch_end),
                stimulus_name=epoch_name,
                software=[software_obj],
                script=script_obj,
                stimulus_parameters=[params_obj],
            )
            schema_epochs.append(epoch_obj)

        return schema_epochs

    def extract_whole_session_epoch(self, stim_table: pd.DataFrame) -> tuple[float, float]:
        """
        Extract the overall start and end times for the entire session.

        Parameters
        ----------
        stim_table : pd.DataFrame
            The stimulus table containing start_time and stop_time columns

        Returns
        -------
        tuple[float, float]
            A tuple of (session_start_time, session_end_time)
        """
        return (stim_table["start_time"].iloc[0], stim_table["stop_time"].iloc[-1])



"""Module defining JobSettings for Mesoscope ETL"""

from datetime import datetime
from pathlib import Path
from typing import List, Literal, Optional

from pydantic import Field, field_validator

from aind_metadata_extractor.core import BaseJobSettings


class JobSettings(BaseJobSettings):
    """Data to be entered by the user."""

    job_settings_name: Literal["Mesoscope"] = Field(default="Mesoscope", title="Name of the job settings")
    input_source: Path = Field(..., title="Path to the input source")
    session_id: str = Field(..., title="ID of the session")
    behavior_source: Path = Field(..., title="Path to the behavior source")
    make_camsitm_dir: bool = Field(default=False, title="Make camsitm directory")
    output_directory: Path = Field(..., title="Path to the output directory")
    session_start_time: datetime = Field(..., title="Start time of the session")
    session_end_time: datetime = Field(..., title="End time of the session")
    subject_id: str = Field(..., title="ID of the subject")
    project: str = Field(..., title="Name of the project")
    iacuc_protocol: str = Field(default="2115", title="IACUC protocol number")
    magnification: str = Field(default="16x", title="Magnification")
    fov_coordinate_ml: float = Field(default=1.5, title="Coordinate in ML direction")
    fov_coordinate_ap: float = Field(default=1.5, title="Coordinate in AL direction")
    fov_reference: str = Field(default="Bregma", title="Reference point for the FOV")
    experimenter_full_name: List[str] = Field(title="Full name of the experimenter")
    mouse_platform_name: str = Field(default="disc", title="Name of the mouse platform")
    optional_output: Optional[Path] = Field(default=None, title="Optional output path")

    @field_validator("input_source", "behavior_source", "output_directory")
    @classmethod
    def validate_path_is_dir(cls, v):
        """Validate that the input source is a directory"""
        if not v.is_dir():
            raise ValueError(f"{v} is not a directory")
        return v



"""Generate JSON schema for FIP data model."""

import argparse
import logging
import os
from typing import Annotated, Type

import aind_physiology_fip
from aind_behavior_services.base import SchemaVersionedModel
from aind_behavior_services.schema import export_schema
from aind_physiology_fip.data_mappers import ProtoAcquisitionDataSchema
from pydantic import BaseModel, create_model

__VERSION__ = aind_physiology_fip.__version__

logger = logging.getLogger(__name__)


def write_schema_to_file(file_path: str) -> None:
    """Write the JSON schema to a file."""
    logger.info(f"Writing schema to {file_path}. Using aind-physiology-fip version {__VERSION__}")
    schema = export_schema(_patch_model(ProtoAcquisitionDataSchema), remove_root=False)
    with open(file_path, "w", encoding="utf-8") as f:
        f.write((schema))


def _patch_model(m: Type[BaseModel]) -> Type[BaseModel]:
    """Recursively patch a pydantic model to make version fields unsafe."""
    for field_name, field in m.model_fields.items():
        if isinstance(field.annotation, type) and issubclass(field.annotation, SchemaVersionedModel):
            updates = {}
            if "version" in field.annotation.model_fields:
                updates["version"] = Annotated[str, field.annotation.model_fields["version"]]
            if "aind_behavior_services_pkg_version" in field.annotation.model_fields:
                updates["aind_behavior_services_pkg_version"] = Annotated[
                    str,
                    field.annotation.model_fields["aind_behavior_services_pkg_version"],
                ]
            new_model = create_model(
                f"{field.annotation.__name__}UnsafeVersion",
                __base__=field.annotation,
                __doc__=field.annotation.__doc__,
                **updates,
            )
            m.model_fields[field_name].annotation = new_model
            _patch_model(new_model)
        elif isinstance(field.annotation, type) and issubclass(field.annotation, BaseModel):
            _patch_model(field.annotation)
    return m


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Export schema to JSON file")
    parser.add_argument("--filepath", default=None, help="Path to output JSON file")
    args = parser.parse_args()

    if args.filepath is None:
        script_dir = os.path.dirname(os.path.abspath(__file__))
        filepath = os.path.join(script_dir, "..", "models", "fip.json")
    else:
        filepath = args.filepath

    write_schema_to_file(filepath)
